/* loopopt.c
 *
 * a simple loop optimizer for brainfuck
 *
 * Copyright (C) 2004 Manuel Schiller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>

#include "bf.h"

/* we won't optimize loops referencing more than PTRBUF memory locations
 * note that this limitation is arbitrary and can be removed should the
 * need arise */
#define PTRBUF	128

/* some terminology (there must be better names for these, but I cna't
 * think of any - but hey, you need names for things...):
 * symmetric loops are loops which do not effectively move the pointer
 * primitive loops are loops which do not contains other loops
 * easy loops are loops which do not contain ADD/IN/OUT/ZERO instructions
 *
 * we can only optimize loops which zero the loop counter with decrements 
 * or inrements
 * loops must be symmetric, primitive and easy for this to work
 */

/* find the end of a loop construct and gather some statistics about it */
static long findeol(bfinstr_t *pgm, long ip)
{
	int isasym = 0;		/* non-zero if asymmetric */
	int isnprim = 0;	/* non-zero if not primitive */
	int isneasy = 0;	/* non-zero if not easy */
	int loopcount = 1;

	while (loopcount) {
		switch (pgm[ip].instr) {
		       case PTRI: isasym += pgm[ip].attr; break;
		       case LOOP: isnprim++; loopcount++; break;
		       case POOL: loopcount--; break;
		       case IN:  /* IN/OUT loops should not be optimized */
		       case OUT:
		       case ADD:
		       case ZERO: isneasy++; break;
		       default: break;
		}
		ip++;
	}

	if (isneasy || isasym || isnprim) return -ip;
	return ip;
}

/* this procedure does the real work...
 * it gathers information about which memory locations are modified how
 * and generates a loopless representation, if possible */
static long looprewriter(bfinstr_t *pgm, long ip, long lip)
{
	long ptr[PTRBUF], cnt[PTRBUF];
	long last = 0;
	long cp = 0;
	long tip;
	int tmp;

	for (tmp = 0; tmp < PTRBUF; tmp++) ptr[tmp] = cnt[tmp] = 0;
	
	/* find out what the loop does... */
	for (tip = ip; tip < lip; tip++) {
		switch(pgm[tip].instr) {
			case PTRI:
			     	cp += pgm[tip].attr;
				break;
			case ADDI:
			     	for (tmp = 0; tmp < last; tmp++)
					if (ptr[tmp] == cp) break;
				/* bail out if too many memory locations
				 * are referenced */
				assert (tmp < PTRBUF);
				ptr[tmp] = cp;
				cnt[tmp] += pgm[tip].attr;
				if (tmp == last) last++;
				break;
			default:
				break;
		}
	}

	/* have a look at the loop counter */
     	for (tmp = 0; tmp < last; tmp++)
		if (ptr[tmp] == 0) break;
	/* loop counter is not modified - bail out on endless loops */
	if (tmp == last) return lip;
	if (cnt[tmp] == 0) return lip;
	/* see if it meets our increment/decrement-counter-by-one criterion */
	if ((cnt[tmp] != -1) && (cnt[tmp] != 1)) return lip;
	if (cnt[tmp] == 1) {
		/* this is a kludge to negate the loop counter - the x86
		 * backend will generate a neg, the C backend relies on the
		 * optimizer of the C compiler to do something similar... */
		pgm[ip].instr = ADD;
		pgm[ip].attr = 0;
		pgm[ip++].attr2 = -2;
	}

	/* the combination of ADDI and PTR is always longer than our ADD -
	 * this means we can just overwrite things... */
	for (tmp = 0; tmp < last; tmp++) {
		if (ptr[tmp] == 0) continue;
		if (cnt[tmp] == 0) continue;
		pgm[ip].instr = ADD;
		pgm[ip].attr = ptr[tmp];
		pgm[ip++].attr2 = cnt[tmp];
	}
	/* zeroing the loop counter must be the last thing... */
	pgm[ip].instr = ZERO;
	pgm[ip].attr = 0;
	pgm[ip].attr2 = 0;

	/* remove the gap... */
	tmp = ++ip;
	do {
	   	pgm[tmp++] = pgm[lip];
	} while (pgm[lip++].instr != EOP);

	return ip;
}

/* prototype */
long loopopt(bfinstr_t *pgm, long ip);

/* this tries to optimize a loop */
static long do_loopopt(bfinstr_t *pgm, long ip)
{
	long lip;

	/* try to optimize inner loops away (if there are any) */
	loopopt(pgm, ip);
	lip = findeol(pgm, ip);
	/* we can't do anything if the loop isn't primitive, symmetric, easy */
	if (lip < 0) return -lip;

	return looprewriter(pgm, ip - 1, lip);
}

/* this tries to optimize a code fragment, possibly containing loops */
long loopopt(bfinstr_t *pgm, long ip)
{

	while ((pgm[ip].instr != EOP) && (pgm[ip].instr != POOL))
		if (pgm[ip++].instr == LOOP)
			ip = do_loopopt(pgm, ip);

	return ip;
}
